import json

import psycopg

from . import sql

_JSON_TYPES = {list, dict}


class Postgres:
    def __init__(self, db_url, autocommit=True):
        self._autocommit = autocommit
        self._url = db_url
        self._connect()

    def _commit(self):
        if not self._autocommit:
            self._conn.commit()

    def _connect(self):
        self._conn = psycopg.connect(self._url, autocommit=self._autocommit)

    def _reconnect(self):
        self.close()
        self._connect()

    def close(self):
        self._conn.close()

    def exec(self, query, args=None):
        try:
            with self._conn.cursor() as cur:
                cur.execute(query, args)
            self._commit()
        except Exception as e:
            self._reconnect()
            raise e

    def execs(self, query_args_pairs):
        try:
            with self._conn.cursor() as cur:
                for q, qargs in query_args_pairs:
                    cur.execute(q, qargs)
            self._commit()
        except Exception as e:
            self._reconnect()
            raise e

    def drop(self, *table_names):
        try:
            with self._conn.cursor() as cur:
                for tbl in table_names:
                    cur.execute(sql.drop_q(tbl))
            self._commit()
        except Exception as e:
            self._reconnect()
            raise e

    def create(self, table_name, props):
        with self._conn.cursor() as cur:
            cur.execute(sql.create_q(table_name, props))
            self._commit()
            return True

    def add_column(self, table_name, col):
        with self._conn.cursor() as cur:
            cur.execute(sql.add_column_q(table_name, col))
            self._commit()
            return True

    def index(self, index_name, table_name, columns, unique=False, concurrently=False):
        """
        Create an index. If concurrently=True, uses CONCURRENTLY (requires autocommit).
        """
        if isinstance(columns, str):
            columns = [columns]
        cols = ", ".join(columns)
        uniq = "UNIQUE " if unique else ""
        if concurrently:
            q = f"CREATE {uniq}INDEX CONCURRENTLY IF NOT EXISTS {index_name} ON {table_name} ({cols})"
        else:
            q = sql.index_q(index_name, table_name, columns, unique=unique)
        try:
            with self._conn.cursor() as cur:
                cur.execute(q)
            self._commit()
            return True
        except Exception as e:
            self._reconnect()
            raise e

    def unique(self, index_name, table_name, columns, concurrently=False):
        """
        Ensure a unique index exists on (columns) for table_name.
        Returns True if the index already existed or was created.

        columns: str | list[str]
        """
        if isinstance(columns, str):
            columns = [columns]

        # Check if an equivalent UNIQUE index already exists (any name)
        try:
            with self._conn.cursor() as cur:
                cur.execute(
                    """
                    SELECT indexname, indexdef
                    FROM pg_indexes
                    WHERE tablename = %s
                    """,
                    (table_name,),
                )
                for idxname, idxdef in cur.fetchall():
                    if "UNIQUE" not in idxdef:
                        continue
                    # Very simple parse of column list from "... ON table (col1, col2, ...)"
                    lpar = idxdef.find("(")
                    rpar = idxdef.rfind(")")
                    if lpar != -1 and rpar != -1 and rpar > lpar:
                        def_cols = [
                            c.strip().strip('"')
                            for c in idxdef[lpar + 1 : rpar].split(",")
                        ]
                        if def_cols == columns:
                            return True  # already have a matching UNIQUE index

            # Create it
            if concurrently:
                if not self._autocommit:
                    raise ValueError(
                        "CREATE INDEX CONCURRENTLY requires autocommit=True"
                    )
                q = (
                    f"CREATE UNIQUE INDEX CONCURRENTLY IF NOT EXISTS "
                    f"{index_name} ON {table_name} ({', '.join(columns)})"
                )
            else:
                # Reuse the generic builder
                q = sql.index_q(index_name, table_name, columns, unique=True)

            with self._conn.cursor() as cur:
                cur.execute(q)
                self._commit()
            return True

        except Exception as e:
            self._reconnect()
            raise e

    def select(
        self,
        table_name,
        columns,
        distinct=None,
        where=None,
        order_by=None,
        limit=None,
        join=None,
        offset=None,
        transform=None,
    ):
        try:
            with self._conn.cursor() as cur:
                if columns is None:
                    columns = "*"
                if type(columns) is str:
                    columns = [columns]
                query, args = sql.select_q(
                    table_name,
                    columns,
                    where=where,
                    distinct_on=distinct,
                    order_by=order_by,
                    limit=limit,
                    join=join,
                    offset=offset,
                    placeholder="%s",
                )
                cur.execute(query, args)
                cols = [col.name for col in cur.description]
                res = (dict(zip(cols, vals)) for vals in cur.fetchall())
                if transform is not None:
                    return [transform(el) for el in res]
                return list(res)
        except Exception as e:
            self._reconnect()
            raise e
        finally:
            self._commit()

    def insert(self, table_name, **args):
        global _JSON_TYPES
        returning = args.get("returning", args.get("RETURNING", None))
        try:
            with self._conn.cursor() as cur:
                for k, v in list(args.items()):
                    if k in {"returning", "RETURNING"}:
                        if isinstance(v, str):
                            args[k] = [v]
                    elif type(v) in _JSON_TYPES:
                        args[k] = json.dumps(v)
                args["placeholder"] = "%s"
                q, qargs = sql.insert_q(table_name, **args)
                cur.execute(q, qargs)
                if returning is None:
                    res = None
                else:
                    res = cur.fetchone()
                    if len(res) == 1:
                        res = res[0]
                    else:
                        cols = [col.name for col in cur.description]
                        res = dict(zip(cols, res))
            self._commit()
            return res
        except Exception as e:
            self._reconnect()
            raise e

    def update(self, table_name, bindings, where):
        global _JSON_TYPES
        binds = {"placeholder": "%s", **bindings}
        try:
            with self._conn.cursor() as cur:
                for k, v in list(binds.items()):
                    if type(v) in _JSON_TYPES:
                        binds[k] = json.dumps(v)
                q, args = sql.update_q(table_name, where=where, **binds)
                cur.execute(q, args)
            self._commit()
        except Exception as e:
            self._reconnect()
            raise e

    def delete(self, table_name, where):
        try:
            with self._conn.cursor() as cur:
                cur.execute(*sql.delete_q(table_name, where=where, placeholder="%s"))
            self._commit()
        except Exception as e:
            self._reconnect()
            raise e



import sqlite3
import tempfile
import unittest

from src.pytrivialsql import sqlite


class TestDBInteraction(unittest.TestCase):
    def test_basic_interactions(self):
        with tempfile.NamedTemporaryFile(suffix=".db") as f:
            db = sqlite.Sqlite3(f.name)
            self.assertTrue(
                db.create(
                    "a_table",
                    [
                        "id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL",
                        "a_column TEXT",
                        "a_number_column INTEGER",
                        "a_boolean_column INTEGER",
                        "created DATETIME DEFAULT CURRENT_TIMESTAMP NOT NULL",
                    ],
                )
            )

            # Index creation and verification
            self.assertTrue(db.index("idx_a_table_a_column", "a_table", ["a_column"]))
            self.assertTrue(
                db.index("idx_a_table_a_number_column", "a_table", ["a_number_column"])
            )

            # Introspect indexes directly via PRAGMA
            cur = db._conn.execute("PRAGMA index_list('a_table')")
            idx_names = {row[1] for row in cur.fetchall()}  # row[1] is name
            self.assertIn("idx_a_table_a_column", idx_names)
            self.assertIn("idx_a_table_a_number_column", idx_names)

            rwid = db.insert(
                "a_table", a_column="A Value", a_number_column=42, a_boolean_column=True
            )

            self.assertTrue(
                db.unique(
                    "uniq_a_table_pair", "a_table", ["a_column", "a_number_column"]
                )
            )
            # idempotent
            self.assertTrue(
                db.unique(
                    "uniq_a_table_pair", "a_table", ["a_column", "a_number_column"]
                )
            )

            # Parameterized IN (...) path sanity check at driver level
            rows = db.select("a_table", "*", where={"id": [rwid, -1]})
            self.assertTrue(any(r["id"] == rwid for r in rows))

    def test_sqlite_add_column_idempotent(self):
        with tempfile.NamedTemporaryFile(suffix=".db") as f:
            db = sqlite.Sqlite3(f.name)
            self.assertTrue(db.create("t", ["id INTEGER PRIMARY KEY"]))
            self.assertTrue(db.add_column("t", "a_column TEXT"))
            self.assertTrue(db.add_column("t", "a_column TEXT"))  # second time no-op
            cols = [r[1] for r in db._conn.execute("PRAGMA table_info(t)").fetchall()]
            self.assertIn("a_column", cols)



import re
import sqlite3

from . import sql

_COLNAME_RE = re.compile(r'^\s*(?:[`"\[])?([A-Za-z_][A-Za-z0-9_]*)')


class Sqlite3:
    def __init__(self, db_path):
        self.path = db_path
        self._conn = sqlite3.connect(
            self.path, check_same_thread=not self.is_threadsafe()
        )

    def exec(self, query, args=None):
        with self._conn as cur:
            cur.execute(query, args or ())

    def execs(self, query_args_pairs):
        with self._conn as cur:
            for q, qargs in query_args_pairs:
                cur.execute(q, qargs)

    def is_threadsafe(self):
        mem = sqlite3.connect("file::memory:?cache=shared")
        cur = mem.execute(
            "select * from pragma_compile_options where compile_options like 'THREADSAFE=%'"
        )
        res = cur.fetchall()
        cur.close()
        try:
            return res[0][0].split("=")[1] == "1"
        except Exception:
            return False

    def drop(self, *table_names):
        with self._conn as cur:
            for tbl in table_names:
                cur.execute(sql.drop_q(tbl))

    def create(self, table_name, props):
        try:
            with self._conn as cur:
                cur.execute(sql.create_q(table_name, props))
                return True
        except Exception:
            return False

    def _column_exists(self, table_name, column_name):
        cur = self._conn.execute(f"PRAGMA table_info({table_name})")
        try:
            return any(row[1] == column_name for row in cur.fetchall())  # row[1] = name
        finally:
            cur.close()

    def _extract_colname(self, col_def):
        # Handles:  foo TEXT,  "Foo Bar" TEXT,  `foo` INT,  [foo] TEXT, etc.
        m = _COLNAME_RE.match(col_def)
        return m.group(1) if m else col_def.strip().split()[0]

    def add_column(self, table_name, col_def):
        col_name = self._extract_colname(col_def)
        try:
            if self._column_exists(table_name, col_name):
                return True  # idempotent: column already present
            with self._conn as conn:
                conn.execute(f"ALTER TABLE {table_name} ADD COLUMN {col_def}")
            return True
        except Exception:
            return False

    def index(self, index_name, table_name, columns, unique=False):
        """
        SQLite CREATE INDEX IF NOT EXISTS.
        """
        try:
            with self._conn as cur:
                cur.execute(sql.index_q(index_name, table_name, columns, unique=unique))
            return True
        except Exception:
            return False

    def unique(self, index_name, table_name, columns):
        """
        Ensure a unique index exists on (columns) for table_name.
        Returns True if the index already existed or was created.

        columns: str | list[str]
        """
        if isinstance(columns, str):
            columns = [columns]

        try:
            # 1) Check existing UNIQUE indexes on this table
            cur = self._conn.execute(f"PRAGMA index_list('{table_name}')")
            idx_rows = cur.fetchall()
            cur.close()

            for _, idx_name, is_unique, *_ in idx_rows:
                if not is_unique:
                    continue
                icur = self._conn.execute(f"PRAGMA index_info('{idx_name}')")
                col_rows = icur.fetchall()  # seqno, cid, name
                icur.close()
                existing_cols = [r[2] for r in col_rows]
                if existing_cols == columns:  # exact same column order
                    return True

            # 2) Create the unique index
            with self._conn as conn:
                conn.execute(sql.index_q(index_name, table_name, columns, unique=True))
            return True
        except Exception:
            return False

    def select(
        self,
        table_name,
        columns,
        distinct=None,
        where=None,
        order_by=None,
        limit=None,
        join=None,
        offset=None,
        transform=None,
    ):
        with self._conn as cur:
            c = cur.cursor()
            if columns is None or columns == "*":
                columns = [
                    el[1]
                    for el in c.execute(f"PRAGMA table_info({table_name})").fetchall()
                ]
            if not columns:
                raise Exception(f"No such table {table_name}")
            elif isinstance(columns, str):
                columns = [columns]
            query, args = sql.select_q(
                table_name,
                columns,
                where=where,
                distinct=distinct,
                order_by=order_by,
                join=join,
                limit=limit,
                offset=offset,
            )
            c.execute(query, args)
            res = (dict(zip(columns, vals)) for vals in c.fetchall())
            if transform is not None:
                return [transform(el) for el in res]
            return list(res)

    def insert(self, table_name, **args):
        with self._conn as cur:
            c = cur.cursor()
            c.execute(*sql.insert_q(table_name, **args))
            return c.lastrowid

    def update(self, table_name, bindings, where):
        with self._conn as cur:
            c = cur.cursor()
            q, args = sql.update_q(table_name, where=where, **bindings)
            c.execute(q, args)

    def delete(self, table_name, where):
        with self._conn as cur:
            c = cur.cursor()
            c.execute(*sql.delete_q(table_name, where=where))



def _in_clause_for_seq(col, seq, placeholder):
    vals = list(seq)
    non_null = [v for v in vals if v is not None]
    has_null = any(v is None for v in vals)

    if not non_null and has_null:
        return f"{col} IS NULL", ()

    if not non_null:
        # Empty list, no NULLs: match nothing
        return "1=0", ()

    placeholders = ", ".join([placeholder] * len(non_null))
    base = f"{col} IN ({placeholders})"
    if has_null:
        return f"({base} OR {col} IS NULL)", tuple(non_null)
    return base, tuple(non_null)


def _where_dict_clause_to_string(k, v, placeholder):
    if type(v) in {set, list}:
        return _in_clause_for_seq(k, v, placeholder)
    if type(v) is tuple and len(v) == 2:
        return f"{k} {v[0]} {placeholder}", v[1]
    if v is None:
        return f"{k} IS NULL", None
    return f"{k}={placeholder}", v


def _where_dict_to_string(where, placeholder):
    qstrs = []
    qvars = ()
    for qstr, qvar in (
        _where_dict_clause_to_string(k, v, placeholder) for k, v in where.items()
    ):
        qstrs.append(qstr)
        if qvar is not None:
            # ensure tuple concatenation
            if isinstance(qvar, tuple):
                qvars += qvar
            else:
                qvars += (qvar,)
    return " AND ".join(qstrs), qvars


def _where_arr_to_string(where, placeholder):
    queries = []
    variables = ()
    for w in where:
        q, v = _where_to_string(w, placeholder)
        queries += [f"({q})"]
        variables += v
    return " OR ".join(queries), variables


def _where_and_to_string(where, placeholder):
    qstrs = []
    qvars = ()
    for clause in where[1:]:
        qstr, qvar = _where_to_string(clause, placeholder)
        qstrs.append(qstr)
        qvars += qvar
    return " AND ".join(qstrs), qvars


def _where_tup_to_string(where, placeholder):
    if where[0] == "AND":
        return _where_and_to_string(where, placeholder)
    if len(where) == 3:
        return (f"{where[0]} {where[1]} {placeholder}", (where[2],))
    if len(where) == 2 and where[0] == "NOT":
        qstr, qvar = _where_to_string(where[1], placeholder)
        return f"NOT ({qstr})", qvar


def _where_to_string(where, placeholder):
    if isinstance(where, dict):
        return _where_dict_to_string(where, placeholder)
    if isinstance(where, list):
        return _where_arr_to_string(where, placeholder)
    if isinstance(where, tuple):
        return _where_tup_to_string(where, placeholder)
    return None


def join_to_string(join):
    if len(join) == 4:
        join_type, table, join_from, join_to = join
        return f"{join_type} JOIN {table} ON {join_from} = {join_to}"
    if len(join) == 3:
        table, join_from, join_to = join
        return f" LEFT JOIN {table} ON {join_from} = {join_to}"
    return None


def where_to_string(where, placeholder=None):
    if placeholder is None:
        placeholder = "?"
    res = _where_to_string(where, placeholder)
    if res is not None:
        qstr, qvars = res
        return f" WHERE {qstr}", qvars
    return None


def drop_q(table_name):
    return f"DROP TABLE IF EXISTS {table_name}"


def create_q(table_name, cols):
    return f"CREATE TABLE IF NOT EXISTS {table_name}({', '.join(cols)})"


def add_column_q(table_name, col):
    return f"ALTER TABLE {table_name} ADD COLUMN IF NOT EXISTS {col}"


def index_q(index_name, table_name, columns, unique=False):
    if isinstance(columns, str):
        columns = [columns]
    uniq = "UNIQUE " if unique else ""
    cols = ", ".join(columns)
    return f"CREATE {uniq}INDEX IF NOT EXISTS {index_name} ON {table_name} ({cols})"


def insert_q(table_name, **args):
    placeholder = args.get("placeholder", "?")
    returning = args.get("RETURNING", args.get("returning", None))
    for k in ["placeholder", "RETURNING", "returning"]:
        if k in args:
            del args[k]
    ks = list(args.keys())
    vs = list(args.values())
    ret_clause = f" RETURNING {', '.join(returning)}" if returning is not None else ""
    return (
        f"INSERT INTO {table_name} ({', '.join(ks)}) VALUES ({', '.join([placeholder for _ in vs])}){ret_clause}",
        tuple(vs),
    )


def select_q(
    table_name,
    columns,
    distinct=None,
    distinct_on=None,
    where=None,
    join=None,
    order_by=None,
    limit=None,
    offset=None,
    placeholder=None,
):
    assert (
        distinct is None or distinct_on is None
    ), "Only one of DISTINCT or DISTINCT ON can be passed"
    if placeholder is None:
        placeholder = "?"
    if type(columns) is str:
        columns = [columns]

    d_key = "DISTINCT" if distinct else "DISTINCT ON"
    d_cols = distinct or distinct_on
    dist = ""
    if d_cols:
        if type(d_cols) is str:
            dist = f" {d_key} ({d_cols})"
        else:
            dist = f" {d_key} ({', '.join(d_cols)})"

    query = f"SELECT{dist} {', '.join(columns)} FROM {table_name}"
    args = ()
    if join is not None:
        query += join_to_string(join)
    if where is not None:
        where_str, where_args = where_to_string(where, placeholder)
        query += where_str
        args = where_args
    if order_by is not None:
        order_parts = [part.split(";")[0].strip() for part in order_by.split(",")]
        query += f" ORDER BY {', '.join(order_parts)}"
    if limit is not None:
        query += f" LIMIT {str(limit).split(';')[0]}"
    if offset is not None:
        query += f" OFFSET {str(offset).split(';')[0]}"
    return (query, args)


def update_q(table_name, **kwargs):
    placeholder = kwargs.get("placeholder", "?")
    if "placeholder" in kwargs:
        del kwargs["placeholder"]
    where = kwargs.get("where", None)
    where_str, where_args = ("", ())
    if where is not None:
        del kwargs["where"]
        where_str, where_args = where_to_string(where, placeholder)

    sep = f"={placeholder},"
    query = f"UPDATE {table_name} SET {sep.join(kwargs.keys())}={placeholder}"

    return query + where_str, tuple(kwargs.values()) + where_args


def delete_q(table_name, where, placeholder=None):
    if placeholder is None:
        placeholder = "?"
    where_str, where_args = where_to_string(where, placeholder)
    return f"DELETE FROM {table_name}{where_str}", where_args
